@@ -26,6 +26,10 @@
 #include <cwchar>
 #include <cwctype>
 
+#ifdef _WIN32
+#include <utf8_fwrap.hpp>
+#endif
+
 using namespace std;
 
 Alphabet::Alphabet()
@@ -221,3 +225,55 @@ Alphabet::decode(int const code) const
 {
   return spairinv[code];
 }
+
+void Alphabet::setSymbol(int symbol, wstring newSymbolString) {
+  //Should be a special character!
+  if (symbol < 0) slexicinv[-symbol-1] = newSymbolString;
+}
+
+void
+Alphabet::createLoopbackSymbols(set<int> &symbols, Alphabet &basis, Side s, bool nonTagsToo)
+{
+  // Non-tag letters get the same int in spairinv across alphabets,
+  // but tags may differ, so do those separately afterwards.
+  set<int> tags;
+  for(vector<pair<int, int> >::iterator it = basis.spairinv.begin(),
+                                        limit = basis.spairinv.end();
+      it != limit;
+      it++)
+  {
+    if(s == left) {
+      if(basis.isTag(it->first)) 
+      {
+        tags.insert(it->first);
+      }
+      else if(nonTagsToo)
+      {
+        symbols.insert(operator()(it->first, it->first));
+      }
+    }
+    else {
+      if(basis.isTag(it->second)) 
+      {
+        tags.insert(it->second);
+      }
+      else if(nonTagsToo)
+      {
+        symbols.insert(operator()(it->second, it->second));
+      }
+    }
+  }
+  for(map<wstring, int, Ltstr>::iterator it = basis.slexic.begin(),
+                                         limit = basis.slexic.end();
+      it != limit;
+      it++)
+  {
+    // Only include tags that were actually seen on the correct side
+    if(tags.find(it->second) != tags.end()) 
+    {
+      includeSymbol(it->first);
+      symbols.insert(operator()(operator()(it->first),
+                                operator()(it->first)));
+    }
+  }
+}
