@@ -43,6 +43,26 @@ private:
    */
   xmlTextReaderPtr reader;
   
+  /**
+   * The alt value
+   */
+  wstring alt;
+
+  /**
+   * The variant value (monodix)
+   */
+  wstring variant;
+  
+  /**
+   * The variant value (left side of bidix)
+   */
+  wstring variant_left;
+  
+  /**
+   * The variant value (right side of bidix)
+   */
+  wstring variant_right;
+    
   /**
    * The paradigm being compiled
    */
@@ -64,6 +84,16 @@ private:
    */
   wstring letters;
   
+  /**
+   * Set verbose mode: warnings which may or may not be correct
+   */
+  bool verbose;
+
+  /**
+   * First element (of an entry)
+   */
+  bool first_element;
+
   /**
    * Identifier of all the symbols during the compilation
    */
@@ -200,6 +230,14 @@ private:
    */
   void skip(wstring &name, wstring const &elem);
 
+  /**
+   * Skip all document #text nodes before "elem"
+   * @param name the name of the node
+   * @param elem the name of the expected node
+   * @param open true for open element, false for closed
+   */
+  void skip(wstring &name, wstring const &elem, bool open);
+
   /**
    * Skip all blank #text nodes before "name"
    * @param name the name of the node
@@ -264,10 +302,14 @@ public:
   static wstring const COMPILER_LEMMA_ATTR;
   static wstring const COMPILER_IGNORE_ATTR;
   static wstring const COMPILER_IGNORE_YES_VAL;
+  static wstring const COMPILER_ALT_ATTR;
+  static wstring const COMPILER_V_ATTR;
+  static wstring const COMPILER_VL_ATTR;
+  static wstring const COMPILER_VR_ATTR;
 
 
   /**
-   * Copnstructor
+   * Constructor
    */
   Compiler();
 
@@ -292,6 +334,35 @@ public:
    * @param fd the stream where write the result
    */
   void write(FILE *fd);
+
+  /**
+   * Set verbose output
+   */
+  void setVerbose(bool verbosity = false);
+
+  /**
+   * Set the alt value to use in compilation
+   * @param a the value
+   */
+  void setAltValue(string const &a);
+
+  /**
+   * Set the variant value to use in compilation
+   * @param v the value
+   */
+  void setVariantValue(string const &v);
+
+  /**
+   * Set the variant_left value to use in compilation
+   * @param v the value
+   */
+  void setVariantLeftValue(string const &v);
+
+  /**
+   * Set the variant_right value to use in compilation
+   * @param v the value
+   */
+  void setVariantRightValue(string const &v);
 };
 
 
