@@ -21,9 +21,12 @@
 #include <lttoolbox/exception.h>
 
 #include <iostream>
-
 #include <cerrno>
 
+#ifdef _WIN32
+#include <utf8_fwrap.hpp>
+#endif
+
 using namespace std;
 
 	
@@ -44,21 +47,23 @@ FSTProcessor::FSTProcessor()
   
   caseSensitive = false;
   dictionaryCase = false;
-  compoundDecomposition = false;
+  do_decomposition = false;
   nullFlush = false;
   nullFlushGeneration = false;
+  showControlSymbols = false;
+  biltransSurfaceForms = false;  
+  compoundOnlyLSymbol = 0;
+  compoundRSymbol = 0;
+  compound_max_elements = 4;
 
-  pool = new Pool<vector<int> >(4, vector<int>(50));
-
-  initial_state = new State(pool);
-  current_state = new State(pool);
+  initial_state = new State();
+  current_state = new State();
 }
 
 FSTProcessor::~FSTProcessor()
 {
   delete current_state;
   delete initial_state;
-  delete pool;
 }
 
 void
@@ -408,6 +413,100 @@ FSTProcessor::readGeneration(FILE *input, FILE *output)
   return 0x7fffffff;
 }
 
+pair<wstring, int>
+FSTProcessor::readBilingual(FILE *input, FILE *output)
+{
+  wint_t val = fgetwc_unlocked(input);
+  wstring symbol = L"";
+
+  if(feof(input))
+  {
+    return pair<wstring, int>(symbol, 0x7fffffff);
+  }
+  
+  if(outOfWord)
+  {
+    if(val == L'^')
+    {
+      val = fgetwc_unlocked(input);
+      if(feof(input))
+      {
+        return pair<wstring, int>(symbol, 0x7fffffff);
+      }
+    }
+    else if(val == L'\\')
+    {
+      fputwc_unlocked(val, output);
+      val = fgetwc_unlocked(input);
+      if(feof(input))
+      {
+        return pair<wstring, int>(symbol, 0x7fffffff);
+      }
+      fputwc_unlocked(val,output);
+      skipUntil(input, output, L'^');
+      val = fgetwc_unlocked(input);
+      if(feof(input))
+      {
+        return pair<wstring, int>(symbol, 0x7fffffff);
+      }
+    }
+    else
+    {
+      fputwc_unlocked(val, output);
+      skipUntil(input, output, L'^');
+      val = fgetwc_unlocked(input);
+      if(feof(input))
+      {
+        return pair<wstring, int>(symbol, 0x7fffffff);
+      }
+    }
+    outOfWord = false;
+  }
+
+  if(val == L'\\')
+  {
+    val = fgetwc_unlocked(input);
+    return pair<wstring, int>(symbol, val);
+  }
+  else if(val == L'$')
+  {
+    outOfWord = true;
+    return pair<wstring, int>(symbol, static_cast<int>(L'$'));
+  }
+  else if(val == L'<')
+  {
+    wstring cad = L"";
+    cad += static_cast<wchar_t>(val);
+    while((val = fgetwc_unlocked(input)) != L'>')
+    {
+      if(feof(input))
+      {
+	streamError();
+      }
+      cad += static_cast<wchar_t>(val);
+    }
+    cad += static_cast<wchar_t>(val);
+
+    int res = alphabet(cad);
+    
+    if (res == 0) {
+	    symbol = cad;
+    }
+    return pair<wstring, int>(symbol, res);
+  }
+  else if(val == L'[')
+  {
+    fputws_unlocked(readFullBlock(input, L'[', L']').c_str(), output);
+    return readBilingual(input, output);
+  }
+  else
+  {
+    return pair<wstring, int>(symbol, val);
+  }
+
+  return pair<wstring, int>(symbol, 0x7fffffff);
+}
+
 void
 FSTProcessor::flushBlanks(FILE *output)
 {  
@@ -493,6 +592,27 @@ FSTProcessor::writeEscaped(wstring const &str, FILE *output)
   } 
 }
 
+void
+FSTProcessor::writeEscapedWithTags(wstring const &str, FILE *output)
+{
+  for(unsigned int i = 0, limit = str.size(); i < limit; i++)
+  {    
+    if(str[i] == L'<' && i >=1 && str[i-1] != L'\\')
+    {
+      fputws_unlocked(str.substr(i).c_str(), output);
+      return;
+    }
+
+    if(escaped_chars.find(str[i]) != escaped_chars.end())
+    {
+      fputwc_unlocked(L'\\', output);
+    }
+    fputwc_unlocked(str[i], output);
+  } 
+}
+
+
+
 void
 FSTProcessor::printWord(wstring const &sf, wstring const &lf, FILE *output)
 {
@@ -642,7 +762,86 @@ FSTProcessor::initBiltrans()
   initGeneration();
 }
 
+
 wstring
+FSTProcessor::compoundAnalysis(wstring input_word, bool uppercase, bool firstupper) {
+    const int MAX_COMBINATIONS = 500;
+    //wcerr << L"compoundAnalysis(input_word = " << input_word << L")" << endl;
+
+    State current_state = *initial_state;
+
+    for(unsigned int i=0; i<input_word.size(); i++) {
+        wchar_t val=input_word.at(i);
+
+        //wcerr << val << L" før step " << i << L" current_state = " << current_state.getReadableString(alphabet) << endl;
+        current_state.step_case(val, caseSensitive);
+        
+        if(current_state.size() > MAX_COMBINATIONS) {
+            wcerr << L"Warning: compoundAnalysis's MAX_COMBINATIONS exceeded for '" << input_word << L"'" << endl;
+            wcerr << L"         gave up at char " << i << L" '" << val << L"'." << endl;
+
+            wstring nullString = L"";
+            return  nullString;
+        }
+
+        //wcerr << val << L" eft step " << i << L" current_state = " << current_state.getReadableString(alphabet) << endl;
+
+        if(i < input_word.size()-1)
+            current_state.restartFinals(all_finals, compoundOnlyLSymbol, initial_state, '+');
+        
+        //wcerr << val << " eft rest " << i << " current_state = " << current_state.getReadableString(alphabet) << endl;
+        //wcerr << i << " result = "  << current_state.filterFinals(all_finals, alphabet, escaped_chars, uppercase, firstupper) << endl;
+        //wcerr << i << " -- size = " << current_state.size() << endl;
+
+        if(current_state.size()==0) {
+            wstring nullString = L"";
+            return nullString;
+        }
+    }
+
+    current_state.pruneCompounds(compoundRSymbol, '+', compound_max_elements);
+    wstring result = current_state.filterFinals(all_finals, alphabet, escaped_chars, uppercase, firstupper);
+    //wcerr << L"rrresult = " << result << endl;
+    
+    return result;
+}
+
+
+
+void 
+FSTProcessor::initDecompositionSymbols() {
+  if ((compoundOnlyLSymbol=alphabet(L"<:co:only-L>")) == 0
+     && (compoundOnlyLSymbol=alphabet(L"<:compound:only-L>")) == 0
+     && (compoundOnlyLSymbol=alphabet(L"<@co:only-L>")) == 0
+     && (compoundOnlyLSymbol=alphabet(L"<@compound:only-L>")) == 0
+     && (compoundOnlyLSymbol=alphabet(L"<compound-only-L>")) == 0)
+  {
+    wcerr << L"Warning: Decomposition symbol <:compound:only-L> not found" << endl;
+  }
+  else if (!showControlSymbols)
+      alphabet.setSymbol(compoundOnlyLSymbol, L"");
+
+  if ((compoundRSymbol=alphabet(L"<:co:R>")) == 0
+     && (compoundRSymbol=alphabet(L"<:compound:R>")) == 0
+     && (compoundRSymbol=alphabet(L"<@co:R>")) == 0
+     && (compoundRSymbol=alphabet(L"<@compound:R>")) == 0
+     && (compoundRSymbol=alphabet(L"<compound-R>")) == 0) 
+  {
+    wcerr << L"Warning: Decomposition symbol <:compound:R> not found" << endl;
+  }
+  else if (!showControlSymbols)
+      alphabet.setSymbol(compoundRSymbol, L"");
+}
+
+
+void 
+FSTProcessor::initDecomposition() {
+  do_decomposition = true;
+  initAnalysis();
+  initDecompositionSymbols();
+}
+
+/*wstring
 FSTProcessor::decompose(wstring w) 
 {
         State current_state = *initial_state;
@@ -807,7 +1006,7 @@ FSTProcessor::decompose(wstring w)
         }
         //wcerr << L"+ decompose: " << lf << endl;
         return lf;
-}
+}*/
 
 void
 FSTProcessor::analysis(FILE *input, FILE *output)
@@ -839,6 +1038,10 @@ FSTProcessor::analysis(FILE *input, FILE *output)
           uppercase = firstupper && iswupper(sf[sf.size()-1]);
         }
 
+        if(do_decomposition && compoundOnlyLSymbol != 0) 
+        {
+          current_state.pruneStatesWithForbiddenSymbol(compoundOnlyLSymbol);
+        }
         lf = current_state.filterFinals(all_finals, alphabet,
                                         escaped_chars,
                                         uppercase, firstupper);
@@ -853,6 +1056,10 @@ FSTProcessor::analysis(FILE *input, FILE *output)
           uppercase = firstupper && iswupper(sf[sf.size()-1]);
         }
 
+        if(do_decomposition && compoundOnlyLSymbol != 0) 
+        {
+          current_state.pruneStatesWithForbiddenSymbol(compoundOnlyLSymbol);
+        }
         lf = current_state.filterFinals(all_finals, alphabet,
                                         escaped_chars,
                                         uppercase, firstupper);
@@ -867,6 +1074,10 @@ FSTProcessor::analysis(FILE *input, FILE *output)
           uppercase = firstupper && iswupper(sf[sf.size()-1]);
         }
 
+        if(do_decomposition && compoundOnlyLSymbol != 0) 
+        {
+          current_state.pruneStatesWithForbiddenSymbol(compoundOnlyLSymbol);
+        }
         lf = current_state.filterFinals(all_finals, alphabet,
                                         escaped_chars,
                                         uppercase, firstupper);
@@ -881,6 +1092,10 @@ FSTProcessor::analysis(FILE *input, FILE *output)
           uppercase = firstupper && iswupper(sf[sf.size()-1]);
         }
 
+        if(do_decomposition && compoundOnlyLSymbol != 0) 
+        {
+          current_state.pruneStatesWithForbiddenSymbol(compoundOnlyLSymbol);
+        }
         lf = current_state.filterFinals(all_finals, alphabet, 
                                         escaped_chars, 
                                         uppercase, firstupper);
@@ -969,16 +1184,22 @@ FSTProcessor::analysis(FILE *input, FILE *output)
         if(limit == 0)
         {
           input_buffer.back(sf.size());
-          fputwc_unlocked(sf[0], output);          
+          writeEscaped(sf.substr(0,1), output);          
         }
         else
         { 
           input_buffer.back(1+(size-limit));
           wstring unknown_word = sf.substr(0, limit);
-          if(compoundDecomposition) 
+          if(do_decomposition) 
           {
+            if(!dictionaryCase)
+            {
+              firstupper = iswupper(sf[0]);
+              uppercase = firstupper && iswupper(sf[sf.size()-1]);
+            }
+
             wstring compound = L"";
-            compound = decompose(unknown_word);
+            compound = compoundAnalysis(unknown_word, uppercase, firstupper);
             if(compound != L"") 
             {
               printWord(unknown_word, compound, output);
@@ -1002,16 +1223,22 @@ FSTProcessor::analysis(FILE *input, FILE *output)
         if(limit == 0)
         {
           input_buffer.back(sf.size());
-          fputwc_unlocked(sf[0], output);          
+          writeEscaped(sf.substr(0,1), output);
         }
         else
         { 
           input_buffer.back(1+(size-limit));
           wstring unknown_word = sf.substr(0, limit);
-          if(compoundDecomposition) 
+          if(do_decomposition) 
           {
+            if(!dictionaryCase)
+            {
+              firstupper = iswupper(sf[0]);
+              uppercase = firstupper && iswupper(sf[sf.size()-1]);
+            }
+
             wstring compound = L"";
-            compound = decompose(unknown_word);
+            compound = compoundAnalysis(unknown_word, uppercase, firstupper);
             if(compound != L"") 
             {
               printWord(unknown_word, compound, output);
@@ -1296,19 +1523,27 @@ FSTProcessor::generation(FILE *input, FILE *output, GenerationMode mode)
       fputwc(L'=', output);
       val = readGeneration(input, output);
     }
-    
+
     if(val == L'$' && outOfWord)
     {
       if(sf[0] == L'*' || sf[0] == L'%')
       {
-	if(mode != gm_clean)
+	if(mode != gm_clean && mode != gm_tagged_nm)
         {
 	  writeEscaped(sf, output);
 	}
-	else
+	else if (mode == gm_clean)
 	{
 	  writeEscaped(sf.substr(1), output);
 	}
+	else if(mode == gm_tagged_nm)
+	{
+	  fputwc_unlocked(L'^', output);        
+	  writeEscaped(removeTags(sf.substr(1)), output);
+	  fputwc_unlocked(L'/', output);          
+          writeEscapedWithTags(sf, output);
+	  fputwc_unlocked(L'$', output);
+	}
       }
       else if(sf[0] == L'@')
       {
@@ -1324,13 +1559,25 @@ FSTProcessor::generation(FILE *input, FILE *output, GenerationMode mode)
         {
           writeEscaped(removeTags(sf), output);
         }
+        else if(mode == gm_tagged)
+        {
+          writeEscaped(removeTags(sf), output);
+        }
+        else if(mode == gm_tagged_nm)
+        { 
+	  fputwc_unlocked(L'^', output);        
+	  writeEscaped(removeTags(sf.substr(1)), output);
+	  fputwc_unlocked(L'/', output);          
+          writeEscapedWithTags(sf, output);
+	  fputwc_unlocked(L'$', output);
+        }
       }
       else if(current_state.isFinal(all_finals))
       {
         bool uppercase = sf.size() > 1 && iswupper(sf[1]);
         bool firstupper= iswupper(sf[0]);
 
-        if(mode == gm_tagged)
+        if(mode == gm_tagged || mode == gm_tagged_nm)
         {
 	  fputwc_unlocked(L'^', output);
         }
@@ -1339,10 +1586,10 @@ FSTProcessor::generation(FILE *input, FILE *output, GenerationMode mode)
                                                   escaped_chars,
                                                   uppercase, firstupper).substr(1).c_str(),
 						  output);
-        if(mode == gm_tagged)
+        if(mode == gm_tagged || mode == gm_tagged_nm)
         {
 	  fputwc_unlocked(L'/', output);
-	  fputws_unlocked(sf.c_str(), output);
+	  writeEscapedWithTags(sf, output);
 	  fputwc_unlocked(L'$', output);
         }
 
@@ -1359,10 +1606,27 @@ FSTProcessor::generation(FILE *input, FILE *output, GenerationMode mode)
           writeEscaped(removeTags(sf), output);
         }
         else if(mode == gm_unknown)
+        {
+          if(sf != L"")
+          {
+            fputwc_unlocked(L'#', output);
+            writeEscaped(removeTags(sf), output);
+          }
+        }
+        else if(mode == gm_tagged)
         {
           fputwc_unlocked(L'#', output);
           writeEscaped(removeTags(sf), output);
         }
+        else if(mode == gm_tagged_nm)
+        {
+	  fputwc_unlocked(L'^', output);        
+	  writeEscaped(removeTags(sf), output);
+	  fputwc_unlocked(L'/', output);          
+	  fputwc_unlocked(L'#', output);          
+          writeEscapedWithTags(sf, output);
+	  fputwc_unlocked(L'$', output);
+        }
       }
   
       current_state = *initial_state;
@@ -2033,19 +2297,62 @@ FSTProcessor::bilingual(FILE *input, FILE *output)
   }
 
   State current_state = *initial_state;
-  wstring sf = L"";
-  wstring queue = L"";
-  wstring result = L"";
+  wstring sf = L"";		// source language analysis
+  wstring queue = L"";		// symbols to be added to each target
+  wstring result = L"";		// result of looking up analysis in bidix
   
   outOfWord = false;
  
   skipUntil(input, output, L'^');
-  int val;
+  pair<wstring,int> tr;		// readBilingual return value, containing:
+  int val;			// the alphabet value of current symbol, and
+  wstring symbol = L"";		// the current symbol as a string
+  bool seentags = false;  // have we seen any tags at all in the analysis?
 
-  while((val = readGeneration(input, output)) != 0x7fffffff)
+  bool seensurface = false;
+  wstring surface = L"";
+  
+  while(true)			// ie. while(val != 0x7fffffff)
   {
+    tr = readBilingual(input, output);
+    symbol = tr.first;
+    val = tr.second;
+
+    //fwprintf(stderr, L"> %S : %C : %d\n", tr.first.c_str(), tr.second, tr.second);
+    if(biltransSurfaceForms && !seensurface && !outOfWord) 
+    {
+      while(val != L'/' && val != 0x7fffffff) 
+      {
+        surface = surface + symbol; 
+        alphabet.getSymbol(surface, val);
+        tr = readBilingual(input, output);
+        symbol = tr.first;
+        val = tr.second;
+        //fwprintf(stderr, L" == %S : %C : %d => %S\n", symbol.c_str(), val, val, surface.c_str());
+      }
+      seensurface = true;
+      tr = readBilingual(input, output);
+      symbol = tr.first;
+      val = tr.second;
+    }
+
+    if (val == 0x7fffffff) 
+    {
+      break;
+    }
+    
     if(val == L'$' && outOfWord)
     {
+      if(!seentags)        // if no tags: only return complete matches
+      {
+        bool uppercase = sf.size() > 1 && iswupper(sf[1]);
+        bool firstupper= iswupper(sf[0]);
+
+        result = current_state.filterFinals(all_finals, alphabet,
+                                            escaped_chars,
+                                            uppercase, firstupper, 0);
+      }
+      
       if(sf[0] == L'*')
       {
         printWordBilingual(sf, L"/"+sf, output);
@@ -2055,14 +2362,23 @@ FSTProcessor::bilingual(FILE *input, FILE *output)
         printWordBilingual(sf, compose(result, queue), output);
       }
       else
-      {
-        printWordBilingual(sf, L"/@"+sf, output);
+      { //xxx
+        if(biltransSurfaceForms) 
+        {
+          printWordBilingual(surface, L"/@"+surface, output);
+        }
+        else
+        { 
+          printWordBilingual(sf, L"/@"+sf, output);
+        }
       }
-  
+      seensurface = false;
+      surface = L""; 
       queue = L"";
       result = L"";
       current_state = *initial_state;
       sf = L"";
+      seentags = false;
     }
     else if(iswspace(val) && sf.size() == 0)
     {
@@ -2074,7 +2390,11 @@ FSTProcessor::bilingual(FILE *input, FILE *output)
       {
         sf += L'\\';
       }
-      alphabet.getSymbol(sf, val);
+      alphabet.getSymbol(sf, val); // add symbol to sf iff alphabetic
+      if(val == 0)  // non-alphabetic, possibly unknown tag; add to sf
+      {
+	sf += symbol;
+      }      
     }
     else
     {
@@ -2082,7 +2402,15 @@ FSTProcessor::bilingual(FILE *input, FILE *output)
       {
         sf += L'\\';
       }
-      alphabet.getSymbol(sf,val);
+      alphabet.getSymbol(sf, val); // add symbol to sf iff alphabetic
+      if(val == 0)  // non-alphabetic, possibly unknown tag; add to sf
+      {
+	sf += symbol;
+      }
+      if(alphabet.isTag(val) || val == 0) 
+      {
+        seentags = true;
+      }      
       if(current_state.size() != 0)
       {
 	if(!alphabet.isTag(val) && iswupper(val) && !caseSensitive)
@@ -2105,12 +2433,21 @@ FSTProcessor::bilingual(FILE *input, FILE *output)
       }
       if(current_state.size() == 0 && result != L"")
       {
-        if(alphabet.isTag(val))
+        // We already have a result, but there is still more to read
+        // of the analysis; following tags are not consumed, but
+        // output as target language tags (added to result on
+        // end-of-word)
+        if(alphabet.isTag(val)) // known tag
         {
           alphabet.getSymbol(queue, val);
         }
+        else if (val == 0) // non-alphabetic, possibly unknown tag
+        {
+          queue += symbol;
+        }
         else
         {
+          // There are no more alive transductions and the current symbol is not a tag -- unknown word!
           result = L"";
         }
       }
@@ -2127,6 +2464,7 @@ FSTProcessor::biltransWithQueue(wstring const &input_word, bool with_delim)
   unsigned int end_point = input_word.size()-2;
   wstring queue = L"";
   bool mark = false;
+  bool seentags = false;  // have we seen any tags at all in the analysis?
   
   if(with_delim == false)
   {
@@ -2160,6 +2498,7 @@ FSTProcessor::biltransWithQueue(wstring const &input_word, bool with_delim)
     }
     else if(input_word[i] == L'<')
     {
+      seentags = true;
       symbol = L'<';
       for(unsigned int j = i + 1; j <= end_point; j++)
       {
@@ -2217,27 +2556,46 @@ FSTProcessor::biltransWithQueue(wstring const &input_word, bool with_delim)
     }
     
     if(current_state.size() == 0)
-    { 
+    {
       if(symbol != L"" && result != L"")
       {
         queue.append(symbol);
       }
       else
       {
-	// word is not present
+        // word is not present
         if(with_delim)
-	{
+        {
           result = L"^@" + input_word.substr(1);  
-	}
+        }
         else
-	{
+        {
           result = L"@" + input_word;
-	}
+        }
         return pair<wstring, int>(result, 0);  
       }      
     }
   }
 
+  if (!seentags
+      && L"" == current_state.filterFinals(all_finals, alphabet,
+                                           escaped_chars,
+                                           uppercase, firstupper, 0)) 
+  {
+    // word is not present
+    if(with_delim)
+    {
+      result = L"^@" + input_word.substr(1);  
+    }
+    else
+    {
+      result = L"@" + input_word;
+    }
+    return pair<wstring, int>(result, 0);  
+  }
+        
+
+
   // attach unmatched queue automatically
 
   if(queue != L"")
@@ -2661,10 +3019,11 @@ FSTProcessor::removeTags(wstring const &str)
   return str;
 }
 
+
 void
-FSTProcessor::setDecompoundingMode(bool const value)
+FSTProcessor::setBiltransSurfaceForms(bool const value)
 {
-  compoundDecomposition = value;
+  biltransSurfaceForms = value;
 }
 
 void
@@ -2688,7 +3047,7 @@ FSTProcessor::setNullFlush(bool const value)
 bool
 FSTProcessor::getDecompoundingMode()
 {
-  return compoundDecomposition;
+  return do_decomposition;
 }
 
 bool
