@@ -43,7 +43,8 @@ enum GenerationMode
   gm_clean,      // clear all
   gm_unknown,    // display unknown words, clear transfer and generation tags
   gm_all,        // display all
-  gm_tagged      // tagged generation
+  gm_tagged,     // tagged generation
+  gm_tagged_nm   // clean tagged generation
 };
 
 /**
@@ -57,8 +58,6 @@ private:
    */
   map<wstring, TransExe, Ltstr> transducers;
 
-  Pool<vector<int> > *pool;
-
   /**
    * Current state of lexical analysis
    */
@@ -129,6 +128,12 @@ private:
    */
   bool outOfWord;
 
+  /**
+   * true if we're automatically removing surface forms.
+   */
+  bool biltransSurfaceForms;
+
+
   /**
    * if true, makes always difference between uppercase and lowercase
    * characters
@@ -154,7 +159,28 @@ private:
   /**
    * try analysing unknown words as compounds
    */
-  bool compoundDecomposition;
+  bool do_decomposition;
+
+  /**
+   * Symbol of CompoundOnlyL
+   */
+  int compoundOnlyLSymbol;
+
+  /**
+   * Symbol of CompoundR
+   */
+  int compoundRSymbol;
+
+  /**
+   * Show or not the controls symbols (as compoundRSymbol)
+   */
+   bool showControlSymbols;
+
+  /**
+   * Max compound elements
+   * Hard coded for now, but there might come a switch one day
+   */
+  int compound_max_elements;
 
   /**
    * Prints an error of input stream and exits
@@ -218,6 +244,13 @@ private:
    */
   int readGeneration(FILE *input, FILE *output);
 
+  /**
+   * Read text from stream (biltrans version)
+   * @param input the stream to read
+   * @return the queue of 0-symbols, and the next symbol in the stream
+   */
+  pair<wstring, int> readBilingual(FILE *input, FILE *output);
+
   /**
    * Read text from stream (SAO version)
    * @param input the stream to read
@@ -248,6 +281,16 @@ private:
    */
   void writeEscaped(wstring const &str, FILE *output);
 
+
+  /**
+   * Write a string to an output stream, escaping all escapable characters
+   * but keeping symbols without escaping
+   * @param str the string to write, escaping characters
+   * @param output the stream to write in
+   */
+  void writeEscapedWithTags(wstring const &str, FILE *output);
+
+
   /**
    * Checks if an string ends with a particular suffix
    * @param str the string to test
@@ -287,6 +330,8 @@ private:
    */
   void printUnknownWord(wstring const &sf, FILE *output);
 
+  void initDecompositionSymbols();
+
   vector<wstring> numbers;
   int readTMAnalysis(FILE *input);
 
@@ -294,7 +339,7 @@ private:
   void printSpace(wchar_t const val, FILE *output);
   void skipUntil(FILE *input, FILE *output, wint_t const character);
   static wstring removeTags(wstring const &str);
-  wstring decompose(wstring str);
+  wstring compoundAnalysis(wstring str, bool uppercase, bool firstupper);
   size_t firstNotAlpha(wstring const &sf);
 
   void analysis_wrapper_null_flush(FILE *input, FILE *output);
@@ -338,9 +383,9 @@ public:
 
   void setCaseSensitiveMode(bool const value);
   void setDictionaryCaseMode(bool const value);
+  void setBiltransSurfaceForms(bool const value);
   void setNullFlush(bool const value);
   bool getNullFlush();
-  void setDecompoundingMode(bool const value);
   bool getDecompoundingMode();
 };
 
